@@ -1,7 +1,10 @@
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards   #-}
 
-module Dhall.Import.HTTP where
+module Dhall.Import.HTTP
+    ( fetchFromHttpUrl
+    , Manager
+    ) where
 
 import Control.Monad.IO.Class (MonadIO(..))
 import Control.Monad.Trans.State.Strict (StateT)
@@ -12,10 +15,19 @@ import Data.Semigroup ((<>))
 import qualified Data.Text as Text
 import qualified JavaScript.XHR
 
+import Data.Void (Void)
 import Dhall.Core (URL(..))
 import Dhall.URL (renderURL)
 import Dhall.Import.Types
 
+{-| The GHCJS implementation does not require a `Manager`
+
+    The purpose of this synonym is so that "Dhall.Import.Types" can import a
+    `Manager` type from "Dhall.Import.HTTP" that does the correct thing for
+    both the GHC and GHCJS implementations
+-}
+type Manager = Void
+
 fetchFromHttpUrl
     :: URL
     -> Maybe [(CI ByteString, ByteString)]
