@@ -44,6 +44,7 @@ import Dhall.Syntax
     , ImportMode(..)
     , ImportType(..)
     , MultiLet(..)
+    , PreferAnnotation(..)
     , Scheme(..)
     , URL(..)
     , Var(..)
@@ -54,6 +55,7 @@ import Data.Monoid ((<>))
 import Data.Text (Text)
 import Data.Void (Void, absurd)
 import GHC.Float (double2Float, float2Double)
+import Numeric.Half (fromHalf, toHalf)
 
 import qualified Codec.CBOR.Decoding  as Decoding
 import qualified Codec.CBOR.Encoding  as Encoding
@@ -65,10 +67,10 @@ import qualified Data.ByteString
 import qualified Data.ByteString.Lazy
 import qualified Data.Sequence
 import qualified Data.Text            as Text
-import qualified Dhall.Syntax
 import qualified Dhall.Crypto
 import qualified Dhall.Map
 import qualified Dhall.Set
+import qualified Dhall.Syntax         as Syntax
 import qualified Text.Printf          as Printf
 
 {-| Supported version strings
@@ -306,7 +308,7 @@ decodeExpressionInternal decodeEmbed = go
                                     6  -> return TextAppend
                                     7  -> return ListAppend
                                     8  -> return (Combine Nothing)
-                                    9  -> return Prefer
+                                    9  -> return (Prefer PreferFromSource)
                                     10 -> return CombineTypes
                                     11 -> return ImportAlt
                                     12 -> return Equivalent
@@ -759,7 +761,7 @@ encodeExpressionInternal encodeEmbed = go
         Combine _ l r ->
             encodeOperator 8 l r
 
-        Prefer l r ->
+        Prefer _ l r ->
             encodeOperator 9 l r
 
         CombineTypes l r ->
@@ -874,11 +876,11 @@ encodeExpressionInternal encodeEmbed = go
           where
             n32 = double2Float n64
 
+            n16 = toHalf n32
+
             useFloat = n64 == float2Double n32
 
-            useHalf = n64 == 0.0 || n64 == infinity || n64 == -infinity
-
-            infinity = 1/0 :: Double
+            useHalf = n64 == (float2Double $ fromHalf n16)
 
         -- Fast path for the common case of an uninterpolated string
         TextLit (Chunks [] z) ->
@@ -910,7 +912,7 @@ encodeExpressionInternal encodeEmbed = go
                 : concatMap encodeBinding (toList as) ++ [ go b₁ ]
                 )
           where
-            MultiLet as b₁ = Dhall.Syntax.multiLet a₀ b₀
+            MultiLet as b₁ = Syntax.multiLet a₀ b₀
 
             encodeBinding (Binding _ x _ mA₀ _ a) =
                 [ Encoding.encodeString x
@@ -939,6 +941,9 @@ encodeExpressionInternal encodeEmbed = go
                 (go t)
                 (go _T)
 
+        a@With{} ->
+            go (Syntax.desugarWith a)
+
         Note _ b ->
             go b
 
@@ -1104,7 +1109,7 @@ encodeImport import_ =
                 Nothing ->
                     Encoding.encodeNull
                 Just h ->
-                    encodeExpressionInternal encodeImport (Dhall.Syntax.denote h)
+                    encodeExpressionInternal encodeImport (Syntax.denote h)
 
             scheme₁ = case scheme₀ of
                 HTTP  -> 0
