@@ -57,6 +57,10 @@ import Data.Semigroup (Semigroup(..))
 import Data.Sequence (Seq, ViewL(..), ViewR(..))
 import Data.Text (Text)
 import Data.Void (Void)
+import Dhall.Map (Map)
+import Dhall.Set (Set)
+import GHC.Natural (Natural)
+import Prelude hiding (succ)
 
 import Dhall.Syntax
   ( Binding(..)
@@ -64,14 +68,10 @@ import Dhall.Syntax
   , Chunks(..)
   , Const(..)
   , DhallDouble(..)
+  , PreferAnnotation(..)
   , Var(..)
   )
 
-import Dhall.Map (Map)
-import Dhall.Set (Set)
-import GHC.Natural (Natural)
-import Prelude hiding (succ)
-
 import qualified Data.Char
 import qualified Data.Sequence   as Sequence
 import qualified Data.Set
@@ -452,19 +452,27 @@ eval !env t0 =
         NaturalLit n ->
             VNaturalLit n
         NaturalFold ->
-            VPrim $ \case
-                VNaturalLit n ->
-                    VHLam (Typed "natural" (VConst Type)) $ \natural ->
-                    VHLam (Typed "succ" (natural ~> natural)) $ \succ ->
-                    VHLam (Typed "zero" natural) $ \zero ->
-                    let go !acc 0 = acc
-                        go  acc m = go (vApp succ acc) (m - 1)
-                    in  go zero (fromIntegral n :: Integer)
-                n ->
-                    VPrim $ \natural ->
-                    VPrim $ \succ ->
-                    VPrim $ \zero ->
-                    VNaturalFold n natural succ zero
+            VPrim $ \n ->
+            VPrim $ \natural ->
+            VPrim $ \succ ->
+            VPrim $ \zero ->
+            let inert = VNaturalFold n natural succ zero
+            in  case zero of
+                VPrimVar -> inert
+                _ -> case succ of
+                    VPrimVar -> inert
+                    _ -> case natural of
+                        VPrimVar -> inert
+                        _ -> case n of
+                            VNaturalLit n' ->
+                                -- Use an `Integer` for the loop, due to the
+                                -- following issue:
+                                --
+                                -- https://github.com/ghcjs/ghcjs/issues/782
+                                let go !acc 0 = acc
+                                    go  acc m = go (vApp succ acc) (m - 1)
+                                in  go zero (fromIntegral n' :: Integer)
+                            _ -> inert
         NaturalBuild ->
             VPrim $ \case
                 VPrimVar ->
@@ -495,7 +503,12 @@ eval !env t0 =
             x@(VNaturalLit m) ->
                 VPrim $ \case
                     VNaturalLit n
-                        | n >= m    -> VNaturalLit (subtract m n)
+                        | n >= m ->
+                            -- Use an `Integer` for the subtraction, due to the
+                            -- following issue:
+                            --
+                            -- https://github.com/ghcjs/ghcjs/issues/782
+                            VNaturalLit (fromIntegral (subtract (fromIntegral m :: Integer) (fromIntegral n :: Integer)))
                         | otherwise -> VNaturalLit 0
                     y -> VNaturalSubtract x y
             x ->
@@ -580,17 +593,23 @@ eval !env t0 =
 
         ListFold ->
             VPrim $ \a ->
-            VPrim $ \case
-                VListLit _ as ->
-                    VHLam (Typed "list" (VConst Type)) $ \list ->
-                    VHLam (Typed "cons" (a ~> list ~> list) ) $ \cons ->
-                    VHLam (Typed "nil"  list) $ \nil ->
-                    foldr' (\x b -> cons `vApp` x `vApp` b) nil as
-                as ->
-                    VPrim $ \t ->
-                    VPrim $ \c ->
-                    VPrim $ \n ->
-                    VListFold a as t c n
+            VPrim $ \as ->
+            VPrim $ \list ->
+            VPrim $ \cons ->
+            VPrim $ \nil ->
+            let inert = VListFold a as list cons nil
+            in  case nil of
+                VPrimVar -> inert
+                _ -> case cons of
+                    VPrimVar -> inert
+                    _ -> case list of
+                        VPrimVar -> inert
+                        _ -> case a of
+                            VPrimVar -> inert
+                            _ -> case as of
+                                VListLit _ as' ->
+                                    foldr' (\x b -> cons `vApp` x `vApp` b) nil as'
+                                _ -> inert
         ListLength ->
             VPrim $ \ a ->
             VPrim $ \case
@@ -688,10 +707,10 @@ eval !env t0 =
             vCombine mk (eval env t) (eval env u)
         CombineTypes t u ->
             vCombineTypes (eval env t) (eval env u)
-        Prefer t u ->
+        Prefer _ t u ->
             vPrefer env (eval env t) (eval env u)
         RecordCompletion t u ->
-            eval env (Annot (Prefer (Field t "default") u) (Field t "Type"))
+            eval env (Annot (Prefer PreferFromCompletion (Field t "default") u) (Field t "Type"))
         Merge x y ma ->
             case (eval env x, eval env y, fmap (eval env) ma) of
                 (VRecordLit m, VInject _ k mt, _)
@@ -736,6 +755,8 @@ eval !env t0 =
             VAssert (eval env t)
         Equivalent t u ->
             VEquivalent (eval env t) (eval env u)
+        e@With{} ->
+            eval env (Syntax.desugarWith e)
         Note _ e ->
             eval env e
         ImportAlt t _ ->
@@ -1117,7 +1138,7 @@ quote !env !t0 =
         VCombineTypes t u ->
             CombineTypes (quote env t) (quote env u)
         VPrefer t u ->
-            Prefer (quote env t) (quote env u)
+            Prefer PreferFromSource (quote env t) (quote env u)
         VMerge t u ma ->
             Merge (quote env t) (quote env u) (fmap (quote env) ma)
         VToMap t ma ->
@@ -1296,8 +1317,8 @@ alphaNormalize = goEnv EmptyNames
                 Combine m (go t) (go u)
             CombineTypes t u ->
                 CombineTypes (go t) (go u)
-            Prefer t u ->
-                Prefer (go t) (go u)
+            Prefer b t u ->
+                Prefer b (go t) (go u)
             RecordCompletion t u ->
                 RecordCompletion (go t) (go u)
             Merge x y ma ->
@@ -1312,6 +1333,8 @@ alphaNormalize = goEnv EmptyNames
                 Assert (go t)
             Equivalent t u ->
                 Equivalent (go t) (go u)
+            With e k v ->
+                With (go e) k (go v)
             Note s e ->
                 Note s (go e)
             ImportAlt t u ->
