@@ -1,5 +1,7 @@
+{-# LANGUAGE CPP               #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards   #-}
+
 {-# OPTIONS_GHC -Wall #-}
 
 module Dhall.Import.Types where
@@ -8,9 +10,10 @@ import Control.Exception (Exception)
 import Control.Monad.Trans.State.Strict (StateT)
 import Data.Dynamic
 import Data.List.NonEmpty (NonEmpty)
-import Data.Map (Map)
+import Dhall.Map (Map)
 import Data.Semigroup ((<>))
 import Data.Text.Prettyprint.Doc (Pretty(..))
+import Data.Void (Void)
 import Dhall.Context (Context)
 import Dhall.Core
   ( Directory (..)
@@ -24,13 +27,16 @@ import Dhall.Core
   , ReifiedNormalizer(..)
   , URL
   )
+#ifdef WITH_HTTP
+import Dhall.Import.Manager (Manager)
+#endif
 import Dhall.Parser (Src)
-import Dhall.TypeCheck (X)
 import Lens.Family (LensLike')
 import System.FilePath (isRelative, splitDirectories)
 
 import qualified Dhall.Context
-import qualified Data.Map      as Map
+import qualified Dhall.Map     as Map
+import qualified Dhall.Substitution
 import qualified Data.Text
 
 -- | A fully 'chained' import, i.e. if it contains a relative path that path is
@@ -38,21 +44,28 @@ import qualified Data.Text
 --   those are well-typed (either of type `List { header : Text, value Text}` or
 --   `List { mapKey : Text, mapValue Text})` and in normal form. These
 --   invariants are preserved by the API exposed by @Dhall.Import@.
-newtype Chained = Chained { chainedImport :: Import }
+newtype Chained = Chained
+    { chainedImport :: Import
+      -- ^ The underlying import
+    }
   deriving (Eq, Ord)
 
 instance Pretty Chained where
     pretty (Chained import_) = pretty import_
 
+-- | An import that has been fully interpeted
 data ImportSemantics = ImportSemantics
-    { importSemantics :: Expr Src X
+    { importSemantics :: Expr Void Void
     -- ^ The fully resolved import, typechecked and beta-normal.
     }
 
 -- | `parent` imports (i.e. depends on) `child`
 data Depends = Depends { parent :: Chained, child :: Chained }
 
-data SemanticCacheMode = IgnoreSemanticCache | UseSemanticCache
+{-| This enables or disables the semantic cache for imports protected by
+    integrity checks
+-}
+data SemanticCacheMode = IgnoreSemanticCache | UseSemanticCache deriving (Eq)
 
 -- | State threaded throughout the import process
 data Status = Status
@@ -68,12 +81,21 @@ data Status = Status
     -- ^ Cache of imported expressions with their node id in order to avoid
     --   importing the same expression twice with different values
 
+#ifdef WITH_HTTP
+    , _manager :: Maybe Manager
+#else
+    , _manager :: Maybe Void
+#endif
+    -- ^ Used to cache the `Manager` when making multiple requests
+
     , _remote :: URL -> StateT Status IO Data.Text.Text
     -- ^ The remote resolver, fetches the content at the given URL.
 
-    , _normalizer :: Maybe (ReifiedNormalizer X)
+    , _substitutions :: Dhall.Substitution.Substitutions Src Void
 
-    , _startingContext :: Context (Expr Src X)
+    , _normalizer :: Maybe (ReifiedNormalizer Void)
+
+    , _startingContext :: Context (Expr Src Void)
 
     , _semanticCacheMode :: SemanticCacheMode
     }
@@ -89,6 +111,10 @@ emptyStatusWith _remote rootDirectory = Status {..}
 
     _cache = Map.empty
 
+    _manager = Nothing
+
+    _substitutions = Dhall.Substitution.empty
+
     _normalizer = Nothing
 
     _startingContext = Dhall.Context.empty
@@ -112,22 +138,34 @@ emptyStatusWith _remote rootDirectory = Status {..}
       , importMode = Code
       }
 
+-- | Lens from a `Status` to its `_stack` field
 stack :: Functor f => LensLike' f Status (NonEmpty Chained)
 stack k s = fmap (\x -> s { _stack = x }) (k (_stack s))
 
+-- | Lens from a `Status` to its `_graph` field
 graph :: Functor f => LensLike' f Status [Depends]
 graph k s = fmap (\x -> s { _graph = x }) (k (_graph s))
 
+-- | Lens from a `Status` to its `_cache` field
 cache :: Functor f => LensLike' f Status (Map Chained ImportSemantics)
 cache k s = fmap (\x -> s { _cache = x }) (k (_cache s))
 
-remote :: Functor f => LensLike' f Status (URL -> StateT Status IO Data.Text.Text)
+-- | Lens from a `Status` to its `_remote` field
+remote
+    :: Functor f => LensLike' f Status (URL -> StateT Status IO Data.Text.Text)
 remote k s = fmap (\x -> s { _remote = x }) (k (_remote s))
 
-normalizer :: Functor f => LensLike' f Status (Maybe (ReifiedNormalizer X))
+-- | Lens from a `Status` to its `_substitutions` field
+substitutions :: Functor f => LensLike' f Status (Dhall.Substitution.Substitutions Src Void)
+substitutions k s =
+    fmap (\x -> s { _substitutions = x }) (k (_substitutions s))
+
+-- | Lens from a `Status` to its `_normalizer` field
+normalizer :: Functor f => LensLike' f Status (Maybe (ReifiedNormalizer Void))
 normalizer k s = fmap (\x -> s {_normalizer = x}) (k (_normalizer s))
 
-startingContext :: Functor f => LensLike' f Status (Context (Expr Src X))
+-- | Lens from a `Status` to its `_startingContext` field
+startingContext :: Functor f => LensLike' f Status (Context (Expr Src Void))
 startingContext k s =
     fmap (\x -> s { _startingContext = x }) (k (_startingContext s))
 
