@@ -159,6 +159,23 @@ data ResolveMode
     | ListTransitiveDependencies
     | ListImmediateDependencies
 
+-- | Groups of subcommands
+data Group
+    = Manipulate
+    | Generate
+    | Interpret
+    | Convert
+    | Miscellaneous
+    | Debugging
+
+groupDescription :: Group -> String
+groupDescription group = case group of
+    Manipulate -> "Manipulate Dhall code"
+    Generate -> "Generate other formats from Dhall"
+    Interpret -> "Interpret Dhall"
+    Convert -> "Convert Dhall to and from its binary representation"
+    Miscellaneous -> "Miscellaneous"
+    Debugging -> "Debugging this interpreter"
 
 -- | `Parser` for the `Options` type
 parseOptions :: Parser Options
@@ -181,12 +198,12 @@ parseOptions =
         f True  = Censor
         f False = NoCensor
 
-subcommand' :: Bool -> String -> String -> Parser a -> Parser a
-subcommand' internal name description parser =
+subcommand :: Group -> String -> String -> Parser a -> Parser a
+subcommand group name description parser =
     Options.Applicative.hsubparser
         (   Options.Applicative.command name parserInfo
         <>  Options.Applicative.metavar name
-        <>  if internal then Options.Applicative.internal else mempty
+        <>  Options.Applicative.commandGroup (groupDescription group)
         )
   where
     parserInfo =
@@ -195,75 +212,85 @@ subcommand' internal name description parser =
             <>  Options.Applicative.progDesc description
             )
 
-subcommand :: String -> String -> Parser a -> Parser a
-subcommand = subcommand' False
-
-internalSubcommand :: String -> String -> Parser a -> Parser a
-internalSubcommand = subcommand' True
-
 parseMode :: Parser Mode
 parseMode =
         subcommand
-            "version"
-            "Display version"
-            (pure Version)
-    <|> subcommand
-            "resolve"
-            "Resolve an expression's imports"
-            (Resolve <$> parseFile <*> parseResolveMode <*> parseSemanticCacheMode)
-    <|> subcommand
-            "type"
-            "Infer an expression's type"
-            (Type <$> parseFile <*> parseQuiet <*> parseSemanticCacheMode)
-    <|> subcommand
-            "normalize"
-            "Normalize an expression"
-            (Normalize <$> parseFile <*> parseAlpha)
-    <|> subcommand
-            "repl"
-            "Interpret expressions in a REPL"
-            (pure Repl)
-    <|> subcommand
-            "diff"
-            "Render the difference between the normal form of two expressions"
-            (Diff <$> argument "expr1" <*> argument "expr2")
-    <|> subcommand
-            "hash"
-            "Compute semantic hashes for Dhall expressions"
-            (Hash <$> parseFile)
-    <|> subcommand
-            "lint"
-            "Improve Dhall code by using newer language features and removing dead code"
-            (Lint <$> parseInplace <*> parseCheck)
-    <|> subcommand
-            "tags"
-            "Generate etags file"
-            (Tags <$> parseInput <*> parseTagsOutput <*> parseSuffixes <*> parseFollowSymlinks)
-    <|> subcommand
+            Manipulate
             "format"
             "Standard code formatter for the Dhall language"
-            (Format <$> parseInplace <*> parseCheck)
+            (Format <$> parseInplace <*> parseCheck "formatted")
     <|> subcommand
+            Manipulate
             "freeze"
             "Add integrity checks to remote import statements of an expression"
-            (Freeze <$> parseInplace <*> parseAllFlag <*> parseCacheFlag <*> parseCheck)
+            (Freeze <$> parseInplace <*> parseAllFlag <*> parseCacheFlag <*> parseCheck "frozen")
     <|> subcommand
-            "encode"
-            "Encode a Dhall expression to binary"
-            (Encode <$> parseFile <*> parseJSONFlag)
-    <|> subcommand
-            "decode"
-            "Decode a Dhall expression from binary"
-            (Decode <$> parseFile <*> parseJSONFlag)
+            Manipulate
+            "lint"
+            "Improve Dhall code by using newer language features and removing dead code"
+            (Lint <$> parseInplace <*> parseCheck "linted")
     <|> subcommand
+            Generate
             "text"
             "Render a Dhall expression that evaluates to a Text literal"
             (Text <$> parseFile)
     <|> subcommand
+            Generate
             "to-directory-tree"
             "Convert nested records of Text literals into a directory tree"
             (DirectoryTree <$> parseFile <*> parseDirectoryTreeOutput)
-    <|> internalSubcommand
+    <|> subcommand
+            Interpret
+            "resolve"
+            "Resolve an expression's imports"
+            (Resolve <$> parseFile <*> parseResolveMode <*> parseSemanticCacheMode)
+    <|> subcommand
+            Interpret
+            "type"
+            "Infer an expression's type"
+            (Type <$> parseFile <*> parseQuiet <*> parseSemanticCacheMode)
+    <|> subcommand
+            Interpret
+            "normalize"
+            "Normalize an expression"
+            (Normalize <$> parseFile <*> parseAlpha)
+    <|> subcommand
+            Convert
+            "encode"
+            "Encode a Dhall expression to binary"
+            (Encode <$> parseFile <*> parseJSONFlag)
+    <|> subcommand
+            Convert
+            "decode"
+            "Decode a Dhall expression from binary"
+            (Decode <$> parseFile <*> parseJSONFlag)
+    <|> subcommand
+            Miscellaneous
+            "repl"
+            "Interpret expressions in a REPL"
+            (pure Repl)
+    <|> subcommand
+            Miscellaneous
+            "diff"
+            "Render the difference between the normal form of two expressions"
+            (Diff <$> argument "expr1" <*> argument "expr2")
+    <|> subcommand
+            Miscellaneous
+            "hash"
+            "Compute semantic hashes for Dhall expressions"
+            (Hash <$> parseFile)
+    <|> subcommand
+            Miscellaneous
+            "tags"
+            "Generate etags file"
+            (Tags <$> parseInput <*> parseTagsOutput <*> parseSuffixes <*> parseFollowSymlinks)
+    <|> subcommand
+            Miscellaneous
+            "version"
+            "Display version"
+            (pure Version)
+    <|> subcommand
+            Debugging
             "haskell-syntax-tree"
             "Output the parsed syntax tree (for debugging)"
             (SyntaxTree <$> parseFile)
@@ -425,7 +452,7 @@ parseMode =
         <>  Options.Applicative.help "Add fallback unprotected imports when using integrity checks purely for caching purposes"
         )
 
-    parseCheck = fmap adapt switch
+    parseCheck processed = fmap adapt switch
       where
         adapt True  = Check
         adapt False = Write
@@ -433,7 +460,7 @@ parseMode =
         switch =
             Options.Applicative.switch
             (   Options.Applicative.long "check"
-            <>  Options.Applicative.help "Only check if the input is formatted"
+            <>  Options.Applicative.help ("Only check if the input is " <> processed)
             )
 
     parseDirectoryTreeOutput =
@@ -734,7 +761,7 @@ command (Options {..}) = do
 
                     let name = case input of
                             InputFile file -> file
-                            StandardInput  -> "(stdin)"
+                            StandardInput  -> "(input)"
 
                     (Header header, expression) <- do
                         Dhall.Core.throws (first Parser.censor (Parser.exprAndHeaderFromText name originalText))
