@@ -24,20 +24,17 @@ import qualified Control.Monad.Fail
 import qualified Data.Char
 import qualified Data.Set
 import qualified Data.Text
-import qualified Data.Text.Prettyprint.Doc               as Pretty
 import qualified Data.Text.Prettyprint.Doc.Render.String as Pretty
 import qualified Dhall.Map
 import qualified Dhall.Pretty
 import qualified Dhall.Set
 import qualified Text.Megaparsec
-#if !MIN_VERSION_megaparsec(7, 0, 0)
-import qualified Text.Megaparsec.Char as Text.Megaparsec (satisfy)
-#endif
 import qualified Text.Megaparsec.Char
 import qualified Text.Parser.Char
 import qualified Text.Parser.Combinators
 import qualified Text.Parser.Token.Style
 
+-- | An exception annotated with a `Src` span
 data SourcedException e = SourcedException Src e
 
 instance Exception e => Exception (SourcedException e)
@@ -48,7 +45,7 @@ instance Show e => Show (SourcedException e) where
         <>  "\n"
         <>  "\n"
         <>  Pretty.renderString
-                (Pretty.layoutPretty Dhall.Pretty.layoutOpts (pretty source))
+                (Dhall.Pretty.layout (pretty source))
 
 -- | Doesn't force the 'Text' part
 laxSrcEq :: Src -> Src -> Bool
@@ -126,9 +123,13 @@ instance MonadPlus Parser where
     -- {-# INLINE mplus #-}
 
 instance Text.Megaparsec.MonadParsec Void Text Parser where
+#if MIN_VERSION_megaparsec(8, 0, 0)
+    parseError e = Parser (Text.Megaparsec.parseError e)
+#else
     failure u e    = Parser (Text.Megaparsec.failure u e)
 
     fancyFailure e = Parser (Text.Megaparsec.fancyFailure e)
+#endif
 
     label l (Parser p) = Parser (Text.Megaparsec.label l p)
 
@@ -196,11 +197,7 @@ instance Text.Parser.Char.CharParsing Parser where
 
   notChar = Text.Megaparsec.Char.char
 
-#if MIN_VERSION_megaparsec(7, 0, 0)
   anyChar = Text.Megaparsec.anySingle
-#else
-  anyChar = Text.Megaparsec.Char.anyChar
-#endif
 
   string = fmap Data.Text.unpack . Text.Megaparsec.Char.string . fromString
 
@@ -260,3 +257,11 @@ toMap kvs = Dhall.Map.unorderedTraverseWithKey (\_k v -> v) m
 
     err k _v1 _v2 = Text.Parser.Combinators.unexpected
                         ("duplicate field: " ++ Data.Text.unpack k)
+
+toMapWith
+    :: (Text -> Parser a -> Parser a -> Parser a)
+    -> [(Text, a)]
+    -> Parser (Map Text a)
+toMapWith combine kvs = sequence m
+  where
+    m = Dhall.Map.fromListWithKey combine (map (\(k, v) -> (k, pure v)) kvs)
