@@ -11,6 +11,7 @@ import Control.Applicative (liftA2, Alternative(..), optional)
 import Data.ByteArray.Encoding (Base(..))
 import Data.Foldable (foldl')
 import Data.Functor (void)
+import Data.List.NonEmpty (NonEmpty(..))
 import Data.Semigroup (Semigroup(..))
 import Data.Text (Text)
 import Dhall.Syntax
@@ -232,7 +233,12 @@ parsers embedded = Parsers {..}
             return (Assert a)
 
         alternative5 = do
-            a <- operatorExpression
+            a0 <- applicationExpression
+
+            let (parseFirstOperatorExpression, parseOperatorExpression) =
+                    operatorExpression (pure a0)
+
+            a <- parseFirstOperatorExpression
 
             let alternative4A = do
                     _arrow
@@ -254,14 +260,63 @@ parsers embedded = Parsers {..}
                             return (ToMap c (Just b))
                         _ -> return (Annot a b)
 
-            alternative4A <|> alternative4B <|> pure a
+            let alternative4C = do
+                    case shallowDenote a of
+                        Equivalent{}   -> empty
+                        ImportAlt{}    -> empty
+                        BoolOr{}       -> empty
+                        NaturalPlus{}  -> empty
+                        TextAppend{}   -> empty
+                        ListAppend{}   -> empty
+                        BoolAnd{}      -> empty
+                        Combine{}      -> empty
+                        Prefer{}       -> empty
+                        CombineTypes{} -> empty
+                        NaturalTimes{} -> empty
+                        BoolEQ{}       -> empty
+                        BoolNE{}       -> empty
+                        App{}          -> empty
+                        _              -> return ()
 
-    operatorExpression =
-        foldr makeOperatorExpression applicationExpression operatorParsers
+                    bs <- many (do
+                        try (whitespace *> _with *> nonemptyWhitespace)
 
-    makeOperatorExpression operatorParser subExpression =
+                        keys <- Combinators.NonEmpty.sepBy1 anyLabel (try (whitespace *> _dot) *> whitespace)
+
+                        whitespace
+
+                        _equal
+
+                        whitespace
+
+                        value <- parseOperatorExpression
+
+                        return (\e -> With e keys value) )
+
+                    return (foldl (\e f -> f e) a0 bs)
+
+            alternative4A <|> alternative4B <|> alternative4C <|> pure a
+
+    -- The firstApplicationExpression argument is necessary in order to
+    -- left-factor the parsers for function types and @with@ expressions to
+    -- minimize backtracking
+    --
+    -- For a longer explanation, see:
+    --
+    -- https://github.com/dhall-lang/dhall-haskell/pull/1770#discussion_r419022486
+    operatorExpression firstApplicationExpression =
+        foldr cons nil operatorParsers
+      where
+        cons operatorParser (p0, p) =
+            ( makeOperatorExpression p0 operatorParser p
+            , makeOperatorExpression p  operatorParser p
+            )
+
+        nil = (firstApplicationExpression, applicationExpression)
+
+    makeOperatorExpression firstSubExpression operatorParser subExpression =
             noted (do
-                a <- subExpression
+                a <- firstSubExpression
 
                 whitespace
 
@@ -277,29 +332,50 @@ parsers embedded = Parsers {..}
 
     operatorParsers :: [Parser (Expr s a -> Expr s a -> Expr s a)]
     operatorParsers =
-        [ ImportAlt       <$ _importAlt    <* nonemptyWhitespace
-        , BoolOr          <$ _or           <* whitespace
-        , NaturalPlus     <$ _plus         <* nonemptyWhitespace
-        , TextAppend      <$ _textAppend   <* whitespace
-        , ListAppend      <$ _listAppend   <* whitespace
-        , BoolAnd         <$ _and          <* whitespace
-        , Combine Nothing <$ _combine      <* whitespace
-        , Prefer          <$ _prefer       <* whitespace
-        , CombineTypes    <$ _combineTypes <* whitespace
-        , NaturalTimes    <$ _times        <* whitespace
-        , BoolEQ          <$ _doubleEqual  <* whitespace
-        , BoolNE          <$ _notEqual     <* whitespace
-        , Equivalent      <$ _equivalent   <* whitespace
+        [ Equivalent              <$ _equivalent   <* whitespace
+        , ImportAlt               <$ _importAlt    <* nonemptyWhitespace
+        , BoolOr                  <$ _or           <* whitespace
+        , NaturalPlus             <$ _plus         <* nonemptyWhitespace
+        , TextAppend              <$ _textAppend   <* whitespace
+        , ListAppend              <$ _listAppend   <* whitespace
+        , BoolAnd                 <$ _and          <* whitespace
+        , Combine Nothing         <$ _combine      <* whitespace
+        , Prefer PreferFromSource <$ _prefer       <* whitespace
+        , CombineTypes            <$ _combineTypes <* whitespace
+        , NaturalTimes            <$ _times        <* whitespace
+        -- Make sure that `==` is not actually the prefix of `===`
+        , BoolEQ                  <$ try (_doubleEqual <* Text.Megaparsec.notFollowedBy (char '=')) <* whitespace
+        , BoolNE                  <$ _notEqual     <* whitespace
         ]
 
     applicationExpression = do
-            f <-    (Some <$ try (_Some <* nonemptyWhitespace))
-                <|> return id
-            a <- noted importExpression_
+            let alternative0 = do
+                    _ <- try (_Some <* nonemptyWhitespace)
+
+                    return (Some, Just "argument to ❰Some❱")
+
+            let alternative1 = do
+                    _ <- try (_toMap *> nonemptyWhitespace)
+
+                    return (\a -> ToMap a Nothing, Just "argument to ❰toMap❱")
+
+            let alternative2 = do
+                    return (id, Nothing)
+
+            (f, maybeMessage) <- alternative0 <|> alternative1 <|> alternative2
+
+            let adapt parser =
+                    case maybeMessage of
+                        Nothing      -> parser
+                        Just message -> parser <?> message
+
+            a <- adapt (noted importExpression_)
+
             bs <- Text.Megaparsec.many . try $ do
                 (sep, _) <- Text.Megaparsec.match nonemptyWhitespace
                 b <- importExpression_
                 return (sep, b)
+
             return (foldl' app (f a) bs)
           where
             app a (sep, b)
@@ -359,7 +435,6 @@ parsers embedded = Parsers {..}
                     , alternative05
                     , alternative06
                     , alternative07
-                    , alternative08
                     , alternative37
                     , alternative09
                     , builtin
@@ -411,11 +486,6 @@ parsers embedded = Parsers {..}
                 b <- importExpression_ <?> "second argument to ❰merge❱"
                 return (Merge a b Nothing)
 
-            alternative08 = do
-                try (_toMap *> nonemptyWhitespace)
-                a <- importExpression_
-                return (ToMap a Nothing)
-
             alternative09 = do
                 a <- try doubleInfinity
                 return (DoubleLit (DhallDouble a))
@@ -730,34 +800,29 @@ parsers embedded = Parsers {..}
             let keysValue = do
                     keys <- Combinators.NonEmpty.sepBy1 anyLabel (try (whitespace *> _dot) *> whitespace)
 
-                    whitespace
+                    let normalRecordEntry = do
+                            try (whitespace *> _equal)
 
-                    _equal
+                            whitespace
 
-                    whitespace
+                            value <- expression
 
-                    value <- expression
+                            let cons key (key', values) =
+                                    (key, RecordLit [ (key', values) ])
 
-                    whitespace
+                            let nil = (NonEmpty.last keys, value)
 
-                    let cons key (key', values) =
-                            (key, RecordLit [ (key', values) ])
+                            return (foldr cons nil (NonEmpty.init keys))
 
-                    let nil = (NonEmpty.last keys, value)
+                    let punnedEntry =
+                            case keys of
+                                x :| [] -> return (x, Var (V x 0))
+                                _       -> empty
 
-                    return (foldr cons nil (NonEmpty.init keys))
+                    (normalRecordEntry <|> punnedEntry) <* whitespace
 
             let nonEmptyRecordLiteral = do
-                    (a, b) <- keysValue
-
-                    e <- Text.Megaparsec.many (do
-                        _comma
-
-                        whitespace
-
-                        (c, d) <- keysValue
-
-                        return (c, d) )
+                    as <- Text.Megaparsec.sepBy1 keysValue (_comma *> whitespace)
 
                     {- The `flip` is necessary because `toMapWith` is internally
                        based on `Data.Map.fromListWithKey` which combines keys
@@ -765,7 +830,7 @@ parsers embedded = Parsers {..}
                     -}
                     let combine k = liftA2 (flip (Combine (Just k)))
 
-                    m <- toMapWith combine ((a, b) : e)
+                    m <- toMapWith combine as
 
                     return (RecordLit m)
 
