@@ -6,7 +6,7 @@ module Dhall.URL where
 import Data.Monoid ((<>))
 import Data.Text (Text)
 
-import Dhall.Core
+import Dhall.Syntax
     ( Scheme(..)
     , URL(..)
     , File(..)
