@@ -7,6 +7,7 @@ import qualified Dhall.Test.Dhall
 import qualified Dhall.Test.Diff
 import qualified Dhall.Test.Tags
 import qualified Dhall.Test.Format
+import qualified Dhall.Test.Freeze
 import qualified Dhall.Test.SemanticHash
 import qualified Dhall.Test.Import
 import qualified Dhall.Test.Lint
@@ -43,6 +44,8 @@ getAllTests = do
 
     semanticHashTests <- Dhall.Test.SemanticHash.getTests
 
+    freezeTests <- Dhall.Test.Freeze.getTests
+
     let testTree =
             Test.Tasty.testGroup "Dhall Tests"
                 [ normalizationTests
@@ -54,6 +57,7 @@ getAllTests = do
                 , diffTests
                 , semanticHashTests
                 , tagsTests
+                , freezeTests
                 , Dhall.Test.Regression.tests
                 , Dhall.Test.Tutorial.tests
                 , Dhall.Test.QuickCheck.tests
@@ -71,6 +75,8 @@ main = do
 
     System.Environment.setEnv "XDG_CACHE_HOME" (pwd </> ".cache")
 
+    System.Environment.setEnv "DHALL_TEST_VAR" "6 * 7"
+
     allTests <- getAllTests
 
     Test.Tasty.defaultMain allTests
