@@ -4,15 +4,12 @@ module Dhall.Test.Parser where
 
 import Data.Monoid ((<>))
 import Data.Text (Text)
+import Data.Void (Void)
 import Dhall.Core (Binding(..), Expr(..), Import, Var(..))
-import Dhall.TypeCheck (X)
 import Prelude hiding (FilePath)
 import Test.Tasty (TestTree)
 import Turtle (FilePath, (</>))
 
-import qualified Codec.CBOR.Read      as CBOR
-import qualified Codec.CBOR.Term      as CBOR
-import qualified Codec.Serialise      as Serialise
 import qualified Control.Monad        as Monad
 import qualified Data.Bifunctor       as Bifunctor
 import qualified Data.ByteString      as ByteString
@@ -26,6 +23,7 @@ import qualified Dhall.Parser         as Parser
 import qualified Dhall.Test.Util      as Test.Util
 import qualified Test.Tasty           as Tasty
 import qualified Test.Tasty.HUnit     as Tasty.HUnit
+import qualified Text.Printf          as Printf
 import qualified Turtle
 
 parseDirectory :: FilePath
@@ -91,7 +89,8 @@ notesInLetInLet = do
                       (Just " ")
                       Nothing
                       (Just " ")
-                      (Note "0 " (NaturalLit 0)))
+                      (Note "0 " (Note "0" (NaturalLit 0)))
+                    )
                     -- This 'Let' isn't wrapped in a 'Note'!
                     (Let
                       (Binding
@@ -100,7 +99,7 @@ notesInLetInLet = do
                         (Just " ")
                         Nothing
                         (Just " ")
-                        (Note "1 " (NaturalLit 1))
+                        (Note "1 " (Note "1" (NaturalLit 1)))
                       )
                       (Note "let z = 2 in x"
                         (Let
@@ -110,10 +109,14 @@ notesInLetInLet = do
                             (Just " ")
                             Nothing
                             (Just " ")
-                            (Note "2 " (NaturalLit 2))
+                            (Note "2 " (Note "2" (NaturalLit 2)))
                           )
-                          (Note "x"
-                            (Var (V "x" 0))))))))
+                          (Note "x" (Var (V "x" 0)))
+                        )
+                      )
+                    )
+                  )
+                )
 
         let msg = "Unexpected parse result"
 
@@ -121,65 +124,45 @@ notesInLetInLet = do
 
 shouldParse :: Text -> TestTree
 shouldParse path = do
-    let skip =
+    let expectedFailures =
             -- This is a bug created by a parsing performance
             -- improvement
-            [ parseDirectory </> "success/unit/MergeParenAnnotation" ]
+            [ parseDirectory </> "success/unit/MergeParenAnnotation"
+            ]
 
     let pathString = Text.unpack path
 
-    Test.Util.testCase path skip $ do
+    Test.Util.testCase path expectedFailures $ do
         text <- Text.IO.readFile (pathString <> "A.dhall")
 
         encoded <- ByteString.Lazy.readFile (pathString <> "B.dhallb")
 
-        expression <- Core.throws (Parser.exprFromText mempty text)
+        expression <- case Parser.exprFromText mempty text of
+            Left  exception  -> Tasty.HUnit.assertFailure (show exception)
+            Right expression -> return expression
 
-        let term = Binary.encodeExpression expression
+        let bytes = Binary.encodeExpression (Core.denote expression)
 
-        let bytes = Serialise.serialise term
+        let render =
+                  concatMap (Printf.printf "%02x ")
+                . ByteString.Lazy.unpack
 
         Monad.unless (encoded == bytes) $ do
-            ("", expected) <- Core.throws (CBOR.deserialiseFromBytes CBOR.decodeTerm encoded)
-
             let message = "The expected CBOR representation doesn't match the actual one\n"
-                          ++ "expected: " ++ show expected ++ "\n but got: " ++ show term
-                          ++ "\n expr: " ++ show expression
+                          ++ "expected: " ++ render encoded ++ "\n but got: " ++ render bytes
+                          ++ "\n expr: " ++ show (Core.denote expression :: Expr Void Import)
 
             Tasty.HUnit.assertFailure message
 
 
 shouldNotParse :: Text -> TestTree
 shouldNotParse path = do
-    let skip =
-            [ -- These two unexpected successes are due to not correctly
-              -- requiring non-empty whitespace after the `:` in a type
-              -- annotation
-              parseDirectory </> "failure/unit/ImportEnvWrongEscape.dhall"
-
-              -- Other spacing related unexpected successes:
-            , parseDirectory </> "failure/spacing/AnnotationNoSpace.dhall"
-            , parseDirectory </> "failure/spacing/ApplicationNoSpace1.dhall"
-            , parseDirectory </> "failure/spacing/ApplicationNoSpace2.dhall"
-            , parseDirectory </> "failure/spacing/AssertNoSpace.dhall"
-            , parseDirectory </> "failure/spacing/ForallNoSpace.dhall"
-            , parseDirectory </> "failure/spacing/ImportAltNoSpace.dhall"
-            , parseDirectory </> "failure/spacing/ImportHashedNoSpace.dhall"
-            , parseDirectory </> "failure/spacing/LambdaNoSpace.dhall"
-            , parseDirectory </> "failure/spacing/ListLitEmptyNoSpace.dhall"
-            , parseDirectory </> "failure/spacing/MergeAnnotationNoSpace3.dhall"
-            , parseDirectory </> "failure/spacing/MergeNoSpace2.dhall"
-            , parseDirectory </> "failure/spacing/NaturalPlusNoSpace.dhall"
-            , parseDirectory </> "failure/spacing/RecordTypeNoSpace.dhall"
-            , parseDirectory </> "failure/spacing/ToMapAnnotNoSpace.dhall"
-            , parseDirectory </> "failure/spacing/UnionTypeNoSpace.dhall"
-
-            , parseDirectory </> "failure/ImportHeadersExteriorHash.dhall"
-
-              -- For parsing performance reasons the implementation
+    let expectedFailures =
+            [ -- For parsing performance reasons the implementation
               -- treats a missing type annotation on an empty list as
               -- as a type-checking failure instead of a parse failure,
               -- but this might be fixable.
+              parseDirectory </> "failure/unit/ListLitEmptyMissingAnnotation.dhall"
             , parseDirectory </> "failure/unit/ListLitEmptyAnnotation.dhall"
 
               -- The same performance improvements also broke the
@@ -189,7 +172,7 @@ shouldNotParse path = do
 
     let pathString = Text.unpack path
 
-    Test.Util.testCase path skip (do
+    Test.Util.testCase path expectedFailures (do
         bytes <- ByteString.readFile pathString
 
         case Text.Encoding.decodeUtf8' bytes of
@@ -201,40 +184,51 @@ shouldNotParse path = do
 
 shouldDecode :: Text -> TestTree
 shouldDecode pathText = do
-    let skip = []
+    let expectedFailures =
+          [ {- Note that this test actually successfully decodes the value, but
+               mistakenly decodes the value to `_` instead of `x`.  This is
+               because the 55799 tag causes normal decoding to fail, so it falls
+               back to treating the `"x"` as a version tag instead of a label.
+
+               Either way, fixing 55799 decoding would cause this test to pass
+               again.
+            -}
+            binaryDecodeDirectory </> "success/unit/SelfDescribeCBORX2"
+          , binaryDecodeDirectory </> "success/unit/SelfDescribeCBORX3"
+          ]
 
     let pathString = Text.unpack pathText
 
-    Test.Util.testCase pathText skip (do
+    Test.Util.testCase pathText expectedFailures (do
         bytes <- ByteString.Lazy.readFile (pathString <> "A.dhallb")
 
-        term <- Core.throws (Serialise.deserialiseOrFail bytes)
-
-        decodedExpression <- Core.throws (Binary.decodeExpression term)
+        decodedExpression <- case Binary.decodeExpression bytes of
+            Left exception ->
+                Tasty.HUnit.assertFailure (show exception)
+            Right decodedExpression ->
+                return decodedExpression
 
         text <- Text.IO.readFile (pathString <> "B.dhall")
 
         parsedExpression <- Core.throws (Parser.exprFromText mempty text)
 
-        let strippedExpression :: Expr X Import
+        let strippedExpression :: Expr Void Import
             strippedExpression = Core.denote parsedExpression
 
         let message =
                 "The decoded expression didn't match the parsed expression"
 
-        Tasty.HUnit.assertEqual message decodedExpression strippedExpression )
+        Tasty.HUnit.assertEqual message strippedExpression decodedExpression )
 
 shouldNotDecode :: Text -> TestTree
 shouldNotDecode pathText = do
-    let skip = []
+    let expectedFailures = []
 
     let pathString = Text.unpack pathText
 
-    Test.Util.testCase pathText skip (do
+    Test.Util.testCase pathText expectedFailures (do
         bytes <- ByteString.Lazy.readFile (pathString <> ".dhallb")
 
-        term <- Core.throws (Serialise.deserialiseOrFail bytes)
-
-        case Binary.decodeExpression term :: Either Binary.DecodingFailure (Expr X Import) of
+        case Binary.decodeExpression bytes :: Either Binary.DecodingFailure (Expr Void Import) of
             Left _ -> return ()
             Right _ -> Tasty.HUnit.assertFailure "Unexpected successful decode" )
