@@ -6,13 +6,14 @@ module Dhall.Test.Tutorial where
 
 import qualified Data.Vector
 import qualified Dhall
-import qualified Dhall.Test.Util as Util
+import qualified Dhall.Test.Substitution as Substitution
+import qualified Dhall.Test.Util         as Util
 import qualified Test.Tasty
 import qualified Test.Tasty.HUnit
 
 import Data.Monoid ((<>))
 import Data.Text (Text)
-import Dhall (Inject)
+import Dhall (ToDhall)
 import GHC.Generics (Generic)
 import Numeric.Natural (Natural)
 import Test.Tasty (TestTree)
@@ -36,6 +37,17 @@ tests =
             , example 3 "./tests/tutorial/unions3A.dhall" "./tests/tutorial/unions3B.dhall"
             , example 4 "./tests/tutorial/unions4A.dhall" "./tests/tutorial/unions4B.dhall"
             ]
+        , Test.Tasty.testGroup "Substitutions"
+            [ Test.Tasty.HUnit.testCase "substitution1.dhall" $ do
+                res <- Substitution.substituteResult "tests/tutorial/substitution1.dhall"
+                res @?= Substitution.Failure 1
+            , Test.Tasty.HUnit.testCase "substitution2.dhall" $ do
+                res <- Substitution.substituteResult "tests/tutorial/substitution2.dhall"
+                res @?= Substitution.Failure 1
+            , Test.Tasty.HUnit.testCase "substitution3.dhall" $ do
+                res <- Substitution.substituteFoo "tests/tutorial/substitution3.dhall"
+                res @?= True
+            ]
         ]
 
 _Interpolation_0 :: TestTree
@@ -69,7 +81,7 @@ _Functions_1 = Test.Tasty.HUnit.testCase "Example #1" (do
     makeBools True False @?= False )
 
 data Example0 = Example0 { foo :: Bool, bar :: Bool }
-    deriving (Generic, Inject)
+    deriving (Generic, ToDhall)
 
 _Functions_2 :: TestTree
 _Functions_2 = Test.Tasty.HUnit.testCase "Example #2" (do
